from typing import Optional
from .serial_com import parse_com, SerialCom 
from pydevmgr_core import BaseNode



class BaseSerialNodeConfig(BaseNode.Config):   
    prefix: str = ""
    # include the possibility to include an optional com configuration for standalone com
    com: Optional[SerialCom.Config] = None
    

class BaseSerialNode(BaseNode):
    Config = BaseSerialNodeConfig
    Com = SerialCom
    def __init__(self, key, config=None, com=None, **kwargs): 
        # parse the config and com object 
            
        super().__init__(key, config=config, com=com, **kwargs)
        self._com = parse_com(com, self._config.com)


    @classmethod
    def new_args(cls, parent, config):
        d = super().new_args(parent, config)
        d.update(com=parent.com)
        return d

    @property
    def com(self):
        return self._com 
    
    @property
    def sid(self):
        self._com.sid
    
    def connect(self):
        self._com.connect()
    
    def disconnect(self):
        self._com.disconnect()
    
    def fget(self):
        raise NotImplementedError('fget')
        
    def fset(self, value):
        raise NotImplementedError('fset')
        



""" lab interface for serial communication """
from pydantic import BaseModel, validator
from enum import Enum
import serial as sr
from typing import Optional, Union 
from pydevmgr_core import kjoin

# Copying serial constant into Enumerator, this will be used by the pydantic model 
class BITSIZE(int, Enum):
    FIVE  = sr.FIVEBITS
    SIX   = sr.SIXBITS
    SEVEN = sr.SEVENBITS
    EIGHT = sr.EIGHTBITS

class PARITY(str, Enum):
    NONE  = sr.PARITY_NONE
    EVEN  = sr.PARITY_EVEN
    ODD   = sr.PARITY_ODD 
    MARK  = sr.PARITY_MARK
    SPACE = sr.PARITY_SPACE

class STOPBITS(float, Enum):
    ONE = sr.STOPBITS_ONE
    ONE_POINT_FIVE = sr.STOPBITS_ONE_POINT_FIVE
    TWO = sr.STOPBITS_TWO



class SerialComConfig(BaseModel):
    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    # Data Structure 
    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    prefix : str = "" # can be used to have a hiearchy in sub communication 
    #----------- from serial.Serial --------------------------------
    port : str = ""
    baudrate: int = 9600
    bytesize:  BITSIZE = BITSIZE.EIGHT
    parity: PARITY = PARITY.NONE
    stopbits: STOPBITS = STOPBITS.ONE
    timeout: Optional[float] = None
    xonxoff: bool = False 
    rtscts: bool = False
    write_timeout: Optional[float] = None 
    inter_byte_timeout: Optional[float] = None
    exclusive: Optional[bool] = None
    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    # Config of BaseModel see pydantic 
    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    class Config:
        use_enum_values = True



class _SerialSubCom:
    """ Wrapper around the :class:`serial.Serial` object 
    
    Providing basic method interface for pydevmgr. 
    """
    def __init__(self, prefix: str, serial: sr.Serial):
        self._prefix = prefix 
        self._serial = serial 
    
    def subcom(self, suffix: str) -> '_SerialSubCom':
        return _SerialSubCom(kjoin(self.prefix, suffix), self._serial)
    
    def connect(self):
        raise RuntimeError("Cannot connect a subcom, use the master com to connect")
    
    def disconnect(self):
        raise RuntimeError("Cannot disconnect a subcom, use the master com to disconnect")        
            
    @property
    def sid(self) -> str:
        """ The port name gives the sid not a big deal here """
        return self._serial.port
    
    @property
    def serial(self) -> sr.Serial:
        return self._serial
    
    @property
    def prefix(self) -> str:
        return self._prefix
    
        
class SerialCom:
    """ Com Wrapper for the serial.Serial object 
    
    The connection (opening of the serial port) is done after connect method is called
    
    Args:
       config (optional, :class:`SerialComConfig`, dict): serial com configuration 
       serial (optional): if a :class:`Serial` object is given it used as is and all other
                          config parameter are ignored. 
                          It is assumed that when a :class:`serial.Serial` object is given 
                          the connection is already opened. 
       **kwargs: If config is None any argument is used to build the Serial config
    """
    Config = SerialComConfig
    BITSIZE = BITSIZE
    PARITY = PARITY
    STOPBITS = STOPBITS
    
    def __init__(self, 
          config: SerialComConfig =None, 
          serial : sr.Serial =None,
          prefix : str = "", 
          **kwargs
         ) -> None:        
        self._is_slave = serial is not None
        
        if serial is None:
            if config is None:
                config = self.Config.parse_obj(kwargs)
            elif isinstance(config, dict):
                config = self.Config.parse_obj({**config, **kwargs})
                
            # open a serial com the port is excluded to avoid connection 
            # the connection is done by the connect method
            serial = sr.Serial(**config.dict(exclude={'port','prefix'}))
            serial.port = config.port
        
        self._prefix = prefix 
        self._config = config # config is not relevant after the the serial has been created
        self._serial = serial 
        
    @property
    def sid(self) -> str:
        """ The port name gives the sid not a big deal here """
        return self._serial.port
    
    @property
    def serial(self) -> sr.Serial:
        return self._serial
    
    @property
    def prefix(self) -> str:
        return self._prefix
    
    @property
    def config(self) -> Config:
        return self._config

    def connect(self) -> None:
        """ open serial communication """
        if self._is_slave:
            raise RuntimeError("Cannot connect connection must be established by the parent com")
        self._serial.open()
    
    def disconnect(self) -> None:
        """ disconnecting serial port """
        if self._is_slave:
            raise RuntimeError("Cannot disconnect connection must be closed by the parent com")
        self._serial.close()
        
    def subcom(self, suffix: str = "") -> _SerialSubCom:
        return _SerialSubCom(kjoin(self.prefix, suffix), self._serial)
        
          

def parse_com(com: Optional[Union[SerialCom, sr.Serial, str]], config: BaseModel) -> SerialCom:
    """ From a com object and a config return a valid :class:`SerialCom` 

    Args:
        com (None, :class:`SerialCom`, :class:`sr.Serial`, str): 
            If string this shall be the port name
        config (BaseModel, dict): Only used if com is None or a string. If com is a string, it overwrite 
            any .com defined in config. The SerialCom will receive a copy of this config
    """
    if com is None:
       com = SerialCom(config=config)            
    elif isinstance(com, (dict, BaseModel)):
        com = SerialCom(config=com)
    elif isinstance(com, sr.Serial):
        com = SerialCom(serial=com)
    elif isinstance(com, str):
        config = dict(config, port=com) 
        com = SerialCom(config=config)     
    return com

    
    
    



from typing import Optional
from .serial_com import parse_com, SerialCom
from pydevmgr_core import BaseDevice



class BaseSerialDeviceConfig(BaseDevice.Config):   
    prefix: str = ""
    # include the possibility to include an optional com configuration for standalone com
    com: SerialCom.Config = SerialCom.Config()
    
    
class BaseSerialDevice(BaseDevice):
    Config = BaseSerialDeviceConfig
    Com = SerialCom
    def __init__(self, 
           key: str, 
           config: Optional[BaseSerialDeviceConfig] = None, 
           com: Optional[SerialCom] = None, 
           **kwargs
        ) -> None:
        # parse config and com 
                 
        super().__init__(key, config=config,  **kwargs) 
        
        self._com = parse_com(com, self._config.com)
    
    
       
    
    @property
    def com(self):
        return self._com 
    
    def connect(self):
        """connect the serial com """
        self._com.connect()
    
    def disconnect(self):
        """disconnect serial com """
        self._com.disconnect()
    
    



from .serial_com import SerialCom
from .serial_device import BaseSerialDevice
from .serial_node import BaseSerialNode
from .serial_interface import BaseSerialInterface
from .serial_rpc import BaseSerialRpc




from typing import Optional 
from .serial_com import parse_com, SerialCom
from pydevmgr_core import BaseInterface



class BaseSerialInterfaceConfig(BaseInterface.Config):   
    prefix: str = ""
    # include the possibility to include an optional com configuration for standalone com
    com: Optional[SerialCom.Config] = None

    
class BaseSerialInterface(BaseInterface):
    Config = BaseSerialInterfaceConfig
    Com = SerialCom
    def __init__(self, 
           key: str, 
           config: Optional[BaseSerialInterfaceConfig] = None, 
           com: Optional[SerialCom] = None, 
           **kwargs
        ) -> None:
        # parse the config and com object                    
        super().__init__(key, config=config, **kwargs) 
        self._com = parse_com(com, self._config.com)

    @property
    def com(self):
        return self._com 
    
    def connect(self):
        """connect the serial com """
        self._com.connect()
    
    def disconnect(self):
        """disconnect serial com """
        self._com.disconnect()
    
    @classmethod
    def new_args(cls, parent, config):
        d = super().new_args(parent, config)
        d.update(com=parent.com)
        return d





from typing import Optional
from .serial_com import SerialCom, parse_com
from pydevmgr_core import BaseRpc



class BaseSerialRpcConfig(BaseRpc.Config):   
    prefix: str = ""
    # include the possibility to include an optional com configuration for standalone com
    com: Optional[SerialCom.Config] = None
    

class BaseSerialRpc(BaseRpc):
    Config = BaseSerialRpcConfig
    Com = SerialCom
    def __init__(self, key, config=None, com=None, **kwargs): 
        # parse the config and com object 
        super().__init__(key, config=config, **kwargs) 
        self._com = parse_com(com, self._config.com)    
    
    @property
    def sid(self):
        self._com.sid
        
    def connect(self):
        self._com.connect()
    
    def disconnect(self):
        self._com.disconnect()
    
    def fcall(self, *args, **kwargs):
        raise NotImplementedError('fcall')
        
        
